@@ -32,13 +32,28 @@ using namespace std;
 
 namespace movit {
 
+namespace {
+
+// An effect whose only purpose is to sit in a phase on its own and take the
+// texture output from a compute shader and display it to the normal backbuffer
+// (or any FBO). That phase can be skipped when rendering using render_to_textures().
+class ComputeShaderOutputDisplayEffect : public Effect {
+public:
+	ComputeShaderOutputDisplayEffect() {}
+	string effect_type_id() const override { return "ComputeShaderOutputDisplayEffect"; }
+	string output_fragment_shader() override { return read_file("identity.frag"); }
+	bool needs_texture_bounce() const override { return true; }
+};
+
+}  // namespace
+
 EffectChain::EffectChain(float aspect_nom, float aspect_denom, ResourcePool *resource_pool)
 	: aspect_nom(aspect_nom),
 	  aspect_denom(aspect_denom),
 	  output_color_rgba(false),
 	  num_output_color_ycbcr(0),
-	  dither_effect(NULL),
-	  ycbcr_conversion_effect_node(NULL),
+	  dither_effect(nullptr),
+	  ycbcr_conversion_effect_node(nullptr),
 	  intermediate_format(GL_RGBA16F),
 	  intermediate_transformation(NO_FRAMEBUFFER_TRANSFORMATION),
 	  num_dither_bits(0),
@@ -46,7 +61,7 @@ EffectChain::EffectChain(float aspect_nom, float aspect_denom, ResourcePool *res
 	  finalized(false),
 	  resource_pool(resource_pool),
 	  do_phase_timing(false) {
-	if (resource_pool == NULL) {
+	if (resource_pool == nullptr) {
 		this->resource_pool = new ResourcePool();
 		owns_resource_pool = true;
 	} else {
@@ -148,8 +163,9 @@ Node *EffectChain::add_node(Effect *effect)
 	node->output_color_space = COLORSPACE_INVALID;
 	node->output_gamma_curve = GAMMA_INVALID;
 	node->output_alpha_type = ALPHA_INVALID;
-	node->needs_mipmaps = false;
+	node->needs_mipmaps = Effect::DOES_NOT_NEED_MIPMAPS;
 	node->one_to_one_sampling = false;
+	node->strong_one_to_one_sampling = false;
 
 	nodes.push_back(node);
 	node_map[effect] = node;
@@ -299,10 +315,10 @@ string replace_prefix(const string &text, const string &prefix)
 namespace {
 
 template<class T>
-void extract_uniform_declarations(const vector<Uniform<T> > &effect_uniforms,
+void extract_uniform_declarations(const vector<Uniform<T>> &effect_uniforms,
                                   const string &type_specifier,
                                   const string &effect_id,
-                                  vector<Uniform<T> > *phase_uniforms,
+                                  vector<Uniform<T>> *phase_uniforms,
                                   string *glsl_string)
 {
 	for (unsigned i = 0; i < effect_uniforms.size(); ++i) {
@@ -315,10 +331,10 @@ void extract_uniform_declarations(const vector<Uniform<T> > &effect_uniforms,
 }
 
 template<class T>
-void extract_uniform_array_declarations(const vector<Uniform<T> > &effect_uniforms,
+void extract_uniform_array_declarations(const vector<Uniform<T>> &effect_uniforms,
                                         const string &type_specifier,
                                         const string &effect_id,
-                                        vector<Uniform<T> > *phase_uniforms,
+                                        vector<Uniform<T>> *phase_uniforms,
                                         string *glsl_string)
 {
 	for (unsigned i = 0; i < effect_uniforms.size(); ++i) {
@@ -335,7 +351,7 @@ void extract_uniform_array_declarations(const vector<Uniform<T> > &effect_unifor
 }
 
 template<class T>
-void collect_uniform_locations(GLuint glsl_program_num, vector<Uniform<T> > *phase_uniforms)
+void collect_uniform_locations(GLuint glsl_program_num, vector<Uniform<T>> *phase_uniforms)
 {
 	for (unsigned i = 0; i < phase_uniforms->size(); ++i) {
 		Uniform<T> &uniform = (*phase_uniforms)[i];
@@ -347,7 +363,12 @@ void collect_uniform_locations(GLuint glsl_program_num, vector<Uniform<T> > *pha
 
 void EffectChain::compile_glsl_program(Phase *phase)
 {
-	string frag_shader_header = read_version_dependent_file("header", "frag");
+	string frag_shader_header;
+	if (phase->is_compute_shader) {
+		frag_shader_header = read_file("header.comp");
+	} else {
+		frag_shader_header = read_version_dependent_file("header", "frag");
+	}
 	string frag_shader = "";
 
 	// Create functions and uniforms for all the texture inputs that we need.
@@ -355,7 +376,7 @@ void EffectChain::compile_glsl_program(Phase *phase)
 		Node *input = phase->inputs[i]->output_node;
 		char effect_id[256];
 		sprintf(effect_id, "in%u", i);
-		phase->effect_ids.insert(make_pair(input, effect_id));
+		phase->effect_ids.insert(make_pair(make_pair(input, IN_ANOTHER_PHASE), effect_id));
 	
 		frag_shader += string("uniform sampler2D tex_") + effect_id + ";\n";
 		frag_shader += string("vec4 ") + effect_id + "(vec2 tc) {\n";
@@ -384,26 +405,43 @@ void EffectChain::compile_glsl_program(Phase *phase)
 		Node *node = phase->effects[i];
 		char effect_id[256];
 		sprintf(effect_id, "eff%u", i);
-		phase->effect_ids.insert(make_pair(node, effect_id));
+		bool inserted = phase->effect_ids.insert(make_pair(make_pair(node, IN_SAME_PHASE), effect_id)).second;
+		assert(inserted);
 	}
 
 	for (unsigned i = 0; i < phase->effects.size(); ++i) {
 		Node *node = phase->effects[i];
-		const string effect_id = phase->effect_ids[node];
-		if (node->incoming_links.size() == 1) {
-			frag_shader += string("#define INPUT ") + phase->effect_ids[node->incoming_links[0]] + "\n";
-		} else {
-			for (unsigned j = 0; j < node->incoming_links.size(); ++j) {
+		const string effect_id = phase->effect_ids[make_pair(node, IN_SAME_PHASE)];
+		for (unsigned j = 0; j < node->incoming_links.size(); ++j) {
+			if (node->incoming_links.size() == 1) {
+				frag_shader += "#define INPUT";
+			} else {
 				char buf[256];
-				sprintf(buf, "#define INPUT%d %s\n", j + 1, phase->effect_ids[node->incoming_links[j]].c_str());
+				sprintf(buf, "#define INPUT%d", j + 1);
 				frag_shader += buf;
 			}
+
+			Node *input = node->incoming_links[j];
+			NodeLinkType link_type = node->incoming_link_type[j];
+			if (i != 0 &&
+			    input->effect->is_compute_shader() &&
+			    node->incoming_link_type[j] == IN_SAME_PHASE) {
+				// First effect after the compute shader reads the value
+				// that cs_output() wrote to a global variable,
+				// ignoring the tc (since all such effects have to be
+				// strong one-to-one).
+				frag_shader += "(tc) CS_OUTPUT_VAL\n";
+			} else {
+				frag_shader += string(" ") + phase->effect_ids[make_pair(input, link_type)] + "\n";
+			}
 		}
 	
 		frag_shader += "\n";
 		frag_shader += string("#define FUNCNAME ") + effect_id + "\n";
+		if (node->effect->is_compute_shader()) {
+			frag_shader += string("#define NORMALIZE_TEXTURE_COORDS(tc) ((tc) * ") + effect_id + "_inv_output_size + " + effect_id + "_output_texcoord_adjust)\n";
+		}
 		frag_shader += replace_prefix(node->effect->output_fragment_shader(), effect_id);
-		frag_shader += "#undef PREFIX\n";
 		frag_shader += "#undef FUNCNAME\n";
 		if (node->incoming_links.size() == 1) {
 			frag_shader += "#undef INPUT\n";
@@ -416,7 +454,17 @@ void EffectChain::compile_glsl_program(Phase *phase)
 		}
 		frag_shader += "\n";
 	}
-	frag_shader += string("#define INPUT ") + phase->effect_ids[phase->effects.back()] + "\n";
+	if (phase->is_compute_shader) {
+		frag_shader += string("#define INPUT ") + phase->effect_ids[make_pair(phase->compute_shader_node, IN_SAME_PHASE)] + "\n";
+		if (phase->compute_shader_node == phase->effects.back()) {
+			// No postprocessing.
+			frag_shader += "#define CS_POSTPROC(tc) CS_OUTPUT_VAL\n";
+		} else {
+			frag_shader += string("#define CS_POSTPROC ") + phase->effect_ids[make_pair(phase->effects.back(), IN_SAME_PHASE)] + "\n";
+		}
+	} else {
+		frag_shader += string("#define INPUT ") + phase->effect_ids[make_pair(phase->effects.back(), IN_SAME_PHASE)] + "\n";
+	}
 
 	// If we're the last phase, add the right #defines for Y'CbCr multi-output as needed.
 	vector<string> frag_shader_outputs;  // In order.
@@ -480,7 +528,14 @@ void EffectChain::compile_glsl_program(Phase *phase)
 		}
 	}
 
-	frag_shader.append(read_file("footer.frag"));
+	if (phase->is_compute_shader) {
+		frag_shader.append(read_file("footer.comp"));
+		phase->compute_shader_node->effect->register_uniform_ivec2("output_size", phase->uniform_output_size);
+		phase->compute_shader_node->effect->register_uniform_vec2("inv_output_size", (float *)&phase->inv_output_size);
+		phase->compute_shader_node->effect->register_uniform_vec2("output_texcoord_adjust", (float *)&phase->output_texcoord_adjust);
+	} else {
+		frag_shader.append(read_file("footer.frag"));
+	}
 
 	// Collect uniforms from all effects and output them. Note that this needs
 	// to happen after output_fragment_shader(), even though the uniforms come
@@ -492,10 +547,12 @@ void EffectChain::compile_glsl_program(Phase *phase)
 	for (unsigned i = 0; i < phase->effects.size(); ++i) {
 		Node *node = phase->effects[i];
 		Effect *effect = node->effect;
-		const string effect_id = phase->effect_ids[node];
+		const string effect_id = phase->effect_ids[make_pair(node, IN_SAME_PHASE)];
+		extract_uniform_declarations(effect->uniforms_image2d, "image2D", effect_id, &phase->uniforms_image2d, &frag_shader_uniforms);
 		extract_uniform_declarations(effect->uniforms_sampler2d, "sampler2D", effect_id, &phase->uniforms_sampler2d, &frag_shader_uniforms);
 		extract_uniform_declarations(effect->uniforms_bool, "bool", effect_id, &phase->uniforms_bool, &frag_shader_uniforms);
 		extract_uniform_declarations(effect->uniforms_int, "int", effect_id, &phase->uniforms_int, &frag_shader_uniforms);
+		extract_uniform_declarations(effect->uniforms_ivec2, "ivec2", effect_id, &phase->uniforms_ivec2, &frag_shader_uniforms);
 		extract_uniform_declarations(effect->uniforms_float, "float", effect_id, &phase->uniforms_float, &frag_shader_uniforms);
 		extract_uniform_declarations(effect->uniforms_vec2, "vec2", effect_id, &phase->uniforms_vec2, &frag_shader_uniforms);
 		extract_uniform_declarations(effect->uniforms_vec3, "vec3", effect_id, &phase->uniforms_vec3, &frag_shader_uniforms);
@@ -507,21 +564,44 @@ void EffectChain::compile_glsl_program(Phase *phase)
 		extract_uniform_declarations(effect->uniforms_mat3, "mat3", effect_id, &phase->uniforms_mat3, &frag_shader_uniforms);
 	}
 
-	frag_shader = frag_shader_header + frag_shader_uniforms + frag_shader;
-
 	string vert_shader = read_version_dependent_file("vs", "vert");
 
 	// If we're the last phase and need to flip the picture to compensate for
-	// the origin, tell the vertex shader so.
-	if (phase->output_node->outgoing_links.empty() && output_origin == OUTPUT_ORIGIN_TOP_LEFT) {
-		const string needle = "#define FLIP_ORIGIN 0";
-		size_t pos = vert_shader.find(needle);
-		assert(pos != string::npos);
+	// the origin, tell the vertex or compute shader so.
+	bool is_last_phase;
+	if (has_dummy_effect) {
+		is_last_phase = (phase->output_node->outgoing_links.size() == 1 &&
+			phase->output_node->outgoing_links[0]->effect->effect_type_id() == "ComputeShaderOutputDisplayEffect");
+	} else {
+		is_last_phase = phase->output_node->outgoing_links.empty();
+	}
+	if (is_last_phase && output_origin == OUTPUT_ORIGIN_TOP_LEFT) {
+		if (phase->is_compute_shader) {
+			frag_shader_header += "#define FLIP_ORIGIN 1\n";
+		} else {
+			const string needle = "#define FLIP_ORIGIN 0";
+			size_t pos = vert_shader.find(needle);
+			assert(pos != string::npos);
 
-		vert_shader[pos + needle.size() - 1] = '1';
+			vert_shader[pos + needle.size() - 1] = '1';
+		}
 	}
 
-	phase->glsl_program_num = resource_pool->compile_glsl_program(vert_shader, frag_shader, frag_shader_outputs);
+	frag_shader = frag_shader_header + frag_shader_uniforms + frag_shader;
+
+	if (phase->is_compute_shader) {
+		phase->glsl_program_num = resource_pool->compile_glsl_compute_program(frag_shader);
+
+		Uniform<int> uniform;
+		uniform.name = "outbuf";
+		uniform.value = &phase->outbuf_image_unit;
+		uniform.prefix = "tex";
+		uniform.num_values = 1;
+		uniform.location = -1;
+		phase->uniforms_image2d.push_back(uniform);
+	} else {
+		phase->glsl_program_num = resource_pool->compile_glsl_program(vert_shader, frag_shader, frag_shader_outputs);
+	}
 	GLint position_attribute_index = glGetAttribLocation(phase->glsl_program_num, "position");
 	GLint texcoord_attribute_index = glGetAttribLocation(phase->glsl_program_num, "texcoord");
 	if (position_attribute_index != -1) {
@@ -532,9 +612,11 @@ void EffectChain::compile_glsl_program(Phase *phase)
 	}
 
 	// Collect the resulting location numbers for each uniform.
+	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_image2d);
 	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_sampler2d);
 	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_bool);
 	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_int);
+	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_ivec2);
 	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_float);
 	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_vec2);
 	collect_uniform_locations(phase->glsl_program_num, &phase->uniforms_vec3);
@@ -558,6 +640,8 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 
 	Phase *phase = new Phase;
 	phase->output_node = output;
+	phase->is_compute_shader = false;
+	phase->compute_shader_node = nullptr;
 
 	// If the output effect has one-to-one sampling, we try to trace this
 	// status down through the dependency chain. This is important in case
@@ -565,6 +649,7 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 	// output size); if we have one-to-one sampling, we don't have to break
 	// the phase.
 	output->one_to_one_sampling = output->effect->one_to_one_sampling();
+	output->strong_one_to_one_sampling = output->effect->strong_one_to_one_sampling();
 
 	// Effects that we have yet to calculate, but that we know should
 	// be in the current phase.
@@ -575,8 +660,14 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 		Node *node = effects_todo_this_phase.top();
 		effects_todo_this_phase.pop();
 
-		if (node->effect->needs_mipmaps()) {
-			node->needs_mipmaps = true;
+		assert(node->effect->one_to_one_sampling() >= node->effect->strong_one_to_one_sampling());
+
+		if (node->effect->needs_mipmaps() != Effect::DOES_NOT_NEED_MIPMAPS) {
+			// Can't have incompatible requirements imposed on us from a dependent effect;
+			// if so, it should have started a new phase instead.
+			assert(node->needs_mipmaps == Effect::DOES_NOT_NEED_MIPMAPS ||
+			       node->needs_mipmaps == node->effect->needs_mipmaps());
+			node->needs_mipmaps = node->effect->needs_mipmaps();
 		}
 
 		// This should currently only happen for effects that are inputs
@@ -591,6 +682,12 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 		}
 
 		phase->effects.push_back(node);
+		if (node->effect->is_compute_shader()) {
+			assert(phase->compute_shader_node == nullptr ||
+			       phase->compute_shader_node == node);
+			phase->is_compute_shader = true;
+			phase->compute_shader_node = node;
+		}
 
 		// Find all the dependencies of this effect, and add them to the stack.
 		vector<Node *> deps = node->incoming_links;
@@ -610,12 +707,29 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 			// Note that we cannot do this propagation as a normal pass,
 			// because it needs information about where the phases end
 			// (we should not propagate the flag across phases).
-			if (node->needs_mipmaps) {
-				if (deps[i]->effect->num_inputs() == 0) {
-					Input *input = static_cast<Input *>(deps[i]->effect);
-					start_new_phase |= !input->can_supply_mipmaps();
-				} else {
-					deps[i]->needs_mipmaps = true;
+			if (node->needs_mipmaps != Effect::DOES_NOT_NEED_MIPMAPS) {
+				// The node can have a value set (ie. not DOES_NOT_NEED_MIPMAPS)
+				// if we have diamonds in the graph; if so, choose that.
+				// If not, the effect on the node can also decide (this is the
+				// more common case).
+				Effect::MipmapRequirements dep_mipmaps = deps[i]->needs_mipmaps;
+				if (dep_mipmaps == Effect::DOES_NOT_NEED_MIPMAPS) {
+					if (deps[i]->effect->num_inputs() == 0) {
+						Input *input = static_cast<Input *>(deps[i]->effect);
+						dep_mipmaps = input->can_supply_mipmaps() ? Effect::DOES_NOT_NEED_MIPMAPS : Effect::CANNOT_ACCEPT_MIPMAPS;
+					} else {
+						dep_mipmaps = deps[i]->effect->needs_mipmaps();
+					}
+				}
+				if (dep_mipmaps == Effect::DOES_NOT_NEED_MIPMAPS) {
+					deps[i]->needs_mipmaps = node->needs_mipmaps;
+				} else if (dep_mipmaps != node->needs_mipmaps) {
+					// The dependency cannot supply our mipmap demands
+					// (either because it's an input that can't do mipmaps,
+					// or because there's a conflict between mipmap-needing
+					// and mipmap-refusing effects somewhere in the graph),
+					// so they cannot be in the same phase.
+					start_new_phase = true;
 				}
 			}
 
@@ -644,7 +758,19 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 				}
 			}
 
-			if (deps[i]->effect->sets_virtual_output_size()) {
+			if (deps[i]->effect->is_compute_shader()) {
+				if (phase->is_compute_shader) {
+					// Only one compute shader per phase.
+					start_new_phase = true;
+				} else if (!node->strong_one_to_one_sampling) {
+					// If all nodes so far are strong one-to-one, we can put them after
+					// the compute shader (ie., process them on the output).
+					start_new_phase = !node->strong_one_to_one_sampling;
+				} else {
+					phase->is_compute_shader = true;
+					phase->compute_shader_node = deps[i];
+				}
+			} else if (deps[i]->effect->sets_virtual_output_size()) {
 				assert(deps[i]->effect->changes_output_size());
 				// If the next effect sets a virtual size to rely on OpenGL's
 				// bilinear sampling, we'll really need to break the phase here.
@@ -663,7 +789,11 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 				// Propagate the one-to-one status down through the dependency.
 				deps[i]->one_to_one_sampling = node->one_to_one_sampling &&
 					deps[i]->effect->one_to_one_sampling();
+				deps[i]->strong_one_to_one_sampling = node->strong_one_to_one_sampling &&
+					deps[i]->effect->strong_one_to_one_sampling();
 			}
+
+			node->incoming_link_type.push_back(start_new_phase ? IN_ANOTHER_PHASE : IN_SAME_PHASE);
 		}
 	}
 
@@ -691,17 +821,13 @@ Phase *EffectChain::construct_phase(Node *output, map<Node *, Phase *> *complete
 	phase->effects = topological_sort(phase->effects);
 
 	// Figure out if we need mipmaps or not, and if so, tell the inputs that.
-	phase->input_needs_mipmaps = false;
-	for (unsigned i = 0; i < phase->effects.size(); ++i) {
-		Node *node = phase->effects[i];
-		phase->input_needs_mipmaps |= node->effect->needs_mipmaps();
-	}
+	// (RTT inputs have different logic, which is checked in execute_phase().)
 	for (unsigned i = 0; i < phase->effects.size(); ++i) {
 		Node *node = phase->effects[i];
 		if (node->effect->num_inputs() == 0) {
 			Input *input = static_cast<Input *>(node->effect);
-			assert(!phase->input_needs_mipmaps || input->can_supply_mipmaps());
-			CHECK(input->set_int("needs_mipmaps", phase->input_needs_mipmaps));
+			assert(node->needs_mipmaps != Effect::NEEDS_MIPMAPS || input->can_supply_mipmaps());
+			CHECK(input->set_int("needs_mipmaps", node->needs_mipmaps == Effect::NEEDS_MIPMAPS));
 		}
 	}
 
@@ -735,7 +861,7 @@ void EffectChain::output_dot(const char *filename)
 	}
 
 	FILE *fp = fopen(filename, "w");
-	if (fp == NULL) {
+	if (fp == nullptr) {
 		perror(filename);
 		exit(1);
 	}
@@ -779,7 +905,7 @@ void EffectChain::output_dot(const char *filename)
 
 		if (nodes[i]->outgoing_links.empty() && !nodes[i]->disabled) {
 			// Output node.
-			vector<string> labels = get_labels_for_edge(nodes[i], NULL);
+			vector<string> labels = get_labels_for_edge(nodes[i], nullptr);
 			output_dot_edge(fp, from_node_id, "output", labels);
 		}
 	}
@@ -792,7 +918,7 @@ vector<string> EffectChain::get_labels_for_edge(const Node *from, const Node *to
 {
 	vector<string> labels;
 
-	if (to != NULL && to->effect->needs_texture_bounce()) {
+	if (to != nullptr && to->effect->needs_texture_bounce()) {
 		labels.push_back("needs_bounce");
 	}
 	if (from->effect->changes_output_size()) {
@@ -959,7 +1085,7 @@ void EffectChain::inform_input_sizes(Phase *phase)
 // desired output size might change based on the inputs.
 void EffectChain::find_output_size(Phase *phase)
 {
-	Node *output_node = phase->effects.back();
+	Node *output_node = phase->is_compute_shader ? phase->compute_shader_node : phase->effects.back();
 
 	// If the last effect explicitly sets an output size, use that.
 	if (output_node->effect->changes_output_size()) {
@@ -1667,6 +1793,44 @@ void EffectChain::add_dither_if_needed()
 	dither_effect = dither->effect;
 }
 
+namespace {
+
+// Whether this effect will cause the phase it is in to become a compute shader phase.
+bool induces_compute_shader(Node *node)
+{
+	if (node->effect->is_compute_shader()) {
+		return true;
+	}
+	if (!node->effect->strong_one_to_one_sampling()) {
+		// This effect can't be chained after a compute shader.
+		return false;
+	}
+	// If at least one of the effects we depend on is a compute shader,
+	// one of them will be put in the same phase as us (the other ones,
+	// if any, will be bounced).
+	for (Node *dep : node->incoming_links) {
+		if (induces_compute_shader(dep)) {
+			return true;
+		}
+	}
+	return false;
+}
+
+}  // namespace
+
+// Compute shaders can't output to the framebuffer, so if the last
+// phase ends in a compute shader, add a dummy phase at the end that
+// only blits directly from the temporary texture.
+void EffectChain::add_dummy_effect_if_needed()
+{
+	Node *output = find_output_node();
+	if (induces_compute_shader(output)) {
+		Node *dummy = add_node(new ComputeShaderOutputDisplayEffect());
+		connect_nodes(output, dummy);
+		has_dummy_effect = true;
+	}
+}
+
 // Find the output node. This is, simply, one that has no outgoing links.
 // If there are multiple ones, the graph is malformed (we do not support
 // multiple outputs right now).
@@ -1738,7 +1902,10 @@ void EffectChain::finalize()
 	output_dot("step18-before-dither.dot");
 	add_dither_if_needed();
 
-	output_dot("step19-final.dot");
+	output_dot("step19-before-dummy-effect.dot");
+	add_dummy_effect_if_needed();
+
+	output_dot("step20-final.dot");
 	
 	// Construct all needed GLSL programs, starting at the output.
 	// We need to keep track of which effects have already been computed,
@@ -1747,7 +1914,7 @@ void EffectChain::finalize()
 	map<Node *, Phase *> completed_effects;
 	construct_phase(find_output_node(), &completed_effects);
 
-	output_dot("step20-split-to-phases.dot");
+	output_dot("step21-split-to-phases.dot");
 
 	assert(phases[0]->inputs.empty());
 	
@@ -1756,18 +1923,6 @@ void EffectChain::finalize()
 
 void EffectChain::render_to_fbo(GLuint dest_fbo, unsigned width, unsigned height)
 {
-	assert(finalized);
-
-	// This needs to be set anew, in case we are coming from a different context
-	// from when we initialized.
-	check_error();
-	glDisable(GL_DITHER);
-	check_error();
-
-	const bool final_srgb = glIsEnabled(GL_FRAMEBUFFER_SRGB);
-	check_error();
-	bool current_srgb = final_srgb;
-
 	// Save original viewport.
 	GLuint x = 0, y = 0;
 
@@ -1780,6 +1935,44 @@ void EffectChain::render_to_fbo(GLuint dest_fbo, unsigned width, unsigned height
 		height = viewport[3];
 	}
 
+	render(dest_fbo, {}, x, y, width, height);
+}
+
+void EffectChain::render_to_texture(const vector<DestinationTexture> &destinations, unsigned width, unsigned height)
+{
+	assert(finalized);
+	assert(!destinations.empty());
+
+	if (!has_dummy_effect) {
+		// We don't end in a compute shader, so there's nothing specific for us to do.
+		// Create an FBO for this set of textures, and just render to that.
+		GLuint texnums[4] = { 0, 0, 0, 0 };
+		for (unsigned i = 0; i < destinations.size() && i < 4; ++i) {
+			texnums[i] = destinations[i].texnum;
+		}
+		GLuint dest_fbo = resource_pool->create_fbo(texnums[0], texnums[1], texnums[2], texnums[3]);
+		render(dest_fbo, {}, 0, 0, width, height);
+		resource_pool->release_fbo(dest_fbo);
+	} else {
+		render((GLuint)-1, destinations, 0, 0, width, height);
+	}
+}
+
+void EffectChain::render(GLuint dest_fbo, const vector<DestinationTexture> &destinations, unsigned x, unsigned y, unsigned width, unsigned height)
+{
+	assert(finalized);
+	assert(destinations.size() <= 1);
+
+	// This needs to be set anew, in case we are coming from a different context
+	// from when we initialized.
+	check_error();
+	glDisable(GL_DITHER);
+	check_error();
+
+	const bool final_srgb = glIsEnabled(GL_FRAMEBUFFER_SRGB);
+	check_error();
+	bool current_srgb = final_srgb;
+
 	// Basic state.
 	check_error();
 	glDisable(GL_BLEND);
@@ -1791,11 +1984,37 @@ void EffectChain::render_to_fbo(GLuint dest_fbo, unsigned width, unsigned height
 
 	set<Phase *> generated_mipmaps;
 
-	// We choose the simplest option of having one texture per output,
-	// since otherwise this turns into an (albeit simple) register allocation problem.
+	// We keep one texture per output, but only for as long as we actually have any
+	// phases that need it as an input. (We don't make any effort to reorder phases
+	// to minimize the number of textures in play, as register allocation can be
+	// complicated and we rarely have much to gain, since our graphs are typically
+	// pretty linear.)
 	map<Phase *, GLuint> output_textures;
+	map<Phase *, int> ref_counts;
+	for (Phase *phase : phases) {
+		for (Phase *input : phase->inputs) {
+			++ref_counts[input];
+		}
+	}
 
-	for (unsigned phase_num = 0; phase_num < phases.size(); ++phase_num) {
+	size_t num_phases = phases.size();
+	if (destinations.empty()) {
+		assert(dest_fbo != (GLuint)-1);
+	} else {
+		assert(has_dummy_effect);
+		assert(x == 0);
+		assert(y == 0);
+		assert(num_phases >= 2);
+		assert(!phases.back()->is_compute_shader);
+		assert(phases.back()->effects.size() == 1);
+		assert(phases.back()->effects[0]->effect->effect_type_id() == "ComputeShaderOutputDisplayEffect");
+
+		// We are rendering to a set of textures, so we can run the compute shader
+		// directly and skip the dummy phase.
+		--num_phases;
+	}
+
+	for (unsigned phase_num = 0; phase_num < num_phases; ++phase_num) {
 		Phase *phase = phases[phase_num];
 
 		if (do_phase_timing) {
@@ -1809,22 +2028,25 @@ void EffectChain::render_to_fbo(GLuint dest_fbo, unsigned width, unsigned height
 			glBeginQuery(GL_TIME_ELAPSED, timer_query_object);
 			phase->timer_query_objects_running.push_back(timer_query_object);
 		}
-		if (phase_num == phases.size() - 1) {
+		bool last_phase = (phase_num == num_phases - 1);
+		if (phase_num == num_phases - 1) {
 			// Last phase goes to the output the user specified.
-			glBindFramebuffer(GL_FRAMEBUFFER, dest_fbo);
-			check_error();
-			GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
-			assert(status == GL_FRAMEBUFFER_COMPLETE);
-			glViewport(x, y, width, height);
-			if (dither_effect != NULL) {
+			if (!phase->is_compute_shader) {
+				glBindFramebuffer(GL_FRAMEBUFFER, dest_fbo);
+				check_error();
+				GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
+				assert(status == GL_FRAMEBUFFER_COMPLETE);
+				glViewport(x, y, width, height);
+			}
+			if (dither_effect != nullptr) {
 				CHECK(dither_effect->set_int("output_width", width));
 				CHECK(dither_effect->set_int("output_height", height));
 			}
 		}
-		bool last_phase = (phase_num == phases.size() - 1);
 
 		// Enable sRGB rendering for intermediates in case we are
 		// rendering to an sRGB format.
+		// TODO: Support this for compute shaders.
 		bool needs_srgb = last_phase ? final_srgb : true;
 		if (needs_srgb && !current_srgb) {
 			glEnable(GL_FRAMEBUFFER_SRGB);
@@ -1836,16 +2058,44 @@ void EffectChain::render_to_fbo(GLuint dest_fbo, unsigned width, unsigned height
 			current_srgb = true;
 		}
 
-		execute_phase(phase, last_phase, &output_textures, &generated_mipmaps);
+		// Find a texture for this phase.
+		inform_input_sizes(phase);
+		find_output_size(phase);
+		vector<DestinationTexture> phase_destinations;
+		if (!last_phase) {
+			GLuint tex_num = resource_pool->create_2d_texture(intermediate_format, phase->output_width, phase->output_height);
+			output_textures.insert(make_pair(phase, tex_num));
+			phase_destinations.push_back(DestinationTexture{ tex_num, intermediate_format });
+
+			// The output texture needs to have valid state to be written to by a compute shader.
+			glActiveTexture(GL_TEXTURE0);
+			check_error();
+			glBindTexture(GL_TEXTURE_2D, tex_num);
+			check_error();
+			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
+			check_error();
+		} else if (phase->is_compute_shader) {
+			assert(!destinations.empty());
+			phase_destinations = destinations;
+		}
+
+		execute_phase(phase, output_textures, phase_destinations, &generated_mipmaps);
 		if (do_phase_timing) {
 			glEndQuery(GL_TIME_ELAPSED);
 		}
+
+		// Drop any input textures we don't need anymore.
+		for (Phase *input : phase->inputs) {
+			assert(ref_counts[input] > 0);
+			if (--ref_counts[input] == 0) {
+				resource_pool->release_2d_texture(output_textures[input]);
+				output_textures.erase(input);
+			}
+		}
 	}
 
-	for (map<Phase *, GLuint>::const_iterator texture_it = output_textures.begin();
-	     texture_it != output_textures.end();
-	     ++texture_it) {
-		resource_pool->release_2d_texture(texture_it->second);
+	for (const auto &phase_and_texnum : output_textures) {
+		resource_pool->release_2d_texture(phase_and_texnum.second);
 	}
 
 	glBindFramebuffer(GL_FRAMEBUFFER, 0);
@@ -1862,8 +2112,8 @@ void EffectChain::render_to_fbo(GLuint dest_fbo, unsigned width, unsigned height
 		// Get back the timer queries.
 		for (unsigned phase_num = 0; phase_num < phases.size(); ++phase_num) {
 			Phase *phase = phases[phase_num];
-			for (std::list<GLuint>::iterator timer_it = phase->timer_query_objects_running.begin();
-			     timer_it != phase->timer_query_objects_running.end(); ) {
+			for (auto timer_it = phase->timer_query_objects_running.cbegin();
+			     timer_it != phase->timer_query_objects_running.cend(); ) {
 				GLint timer_query_object = *timer_it;
 				GLint available;
 				glGetQueryObjectiv(timer_query_object, GL_QUERY_RESULT_AVAILABLE, &available);
@@ -1918,53 +2168,85 @@ void EffectChain::print_phase_timing()
 	printf("Total:   %5.1f ms\n", total_time_ms);
 }
 
-void EffectChain::execute_phase(Phase *phase, bool last_phase,
-                                map<Phase *, GLuint> *output_textures,
+void EffectChain::execute_phase(Phase *phase,
+                                const map<Phase *, GLuint> &output_textures,
+                                const vector<DestinationTexture> &destinations,
                                 set<Phase *> *generated_mipmaps)
 {
-	GLuint fbo = 0;
-
-	// Find a texture for this phase.
-	inform_input_sizes(phase);
-	if (!last_phase) {
-		find_output_size(phase);
-
-		GLuint tex_num = resource_pool->create_2d_texture(intermediate_format, phase->output_width, phase->output_height);
-		output_textures->insert(make_pair(phase, tex_num));
-	}
-
 	// Set up RTT inputs for this phase.
 	for (unsigned sampler = 0; sampler < phase->inputs.size(); ++sampler) {
 		glActiveTexture(GL_TEXTURE0 + sampler);
 		Phase *input = phase->inputs[sampler];
 		input->output_node->bound_sampler_num = sampler;
-		glBindTexture(GL_TEXTURE_2D, (*output_textures)[input]);
+		const auto it = output_textures.find(input);
+		assert(it != output_textures.end());
+		glBindTexture(GL_TEXTURE_2D, it->second);
 		check_error();
-		if (phase->input_needs_mipmaps && generated_mipmaps->count(input) == 0) {
+
+		// See if anything using this RTT input (in this phase) needs mipmaps.
+		// TODO: It could be that we get conflicting logic here, if we have
+		// multiple effects with incompatible mipmaps using the same
+		// RTT input. However, that is obscure enough that we can deal
+		// with it at some future point (preferably when we have
+		// universal support for separate sampler objects!). For now,
+		// an assert is good enough. See also the TODO at bound_sampler_num.
+		bool any_needs_mipmaps = false, any_refuses_mipmaps = false;
+		for (Node *node : phase->effects) {
+			assert(node->incoming_links.size() == node->incoming_link_type.size());
+			for (size_t i = 0; i < node->incoming_links.size(); ++i) {
+				if (node->incoming_links[i] == input->output_node &&
+				    node->incoming_link_type[i] == IN_ANOTHER_PHASE) {
+					if (node->needs_mipmaps == Effect::NEEDS_MIPMAPS) {
+						any_needs_mipmaps = true;
+					} else if (node->needs_mipmaps == Effect::CANNOT_ACCEPT_MIPMAPS) {
+						any_refuses_mipmaps = true;
+					}
+				}
+			}
+		}
+		assert(!(any_needs_mipmaps && any_refuses_mipmaps));
+
+		if (any_needs_mipmaps && generated_mipmaps->count(input) == 0) {
 			glGenerateMipmap(GL_TEXTURE_2D);
 			check_error();
 			generated_mipmaps->insert(input);
 		}
-		setup_rtt_sampler(sampler, phase->input_needs_mipmaps);
+		setup_rtt_sampler(sampler, any_needs_mipmaps);
 		phase->input_samplers[sampler] = sampler;  // Bind the sampler to the right uniform.
 	}
 
-	// And now the output. (Already set up for us if it is the last phase.)
-	if (!last_phase) {
-		fbo = resource_pool->create_fbo((*output_textures)[phase]);
-		glBindFramebuffer(GL_FRAMEBUFFER, fbo);
-		glViewport(0, 0, phase->output_width, phase->output_height);
-	}
-
 	GLuint instance_program_num = resource_pool->use_glsl_program(phase->glsl_program_num);
 	check_error();
 
+	// And now the output.
+	GLuint fbo = 0;
+	if (phase->is_compute_shader) {
+		assert(!destinations.empty());
+
+		// This is currently the only place where we use image units,
+		// so we can always start at 0. TODO: Support multiple destinations.
+		phase->outbuf_image_unit = 0;
+		glBindImageTexture(phase->outbuf_image_unit, destinations[0].texnum, 0, GL_FALSE, 0, GL_WRITE_ONLY, destinations[0].format);
+		check_error();
+		phase->uniform_output_size[0] = phase->output_width;
+		phase->uniform_output_size[1] = phase->output_height;
+		phase->inv_output_size.x = 1.0f / phase->output_width;
+		phase->inv_output_size.y = 1.0f / phase->output_height;
+		phase->output_texcoord_adjust.x = 0.5f / phase->output_width;
+		phase->output_texcoord_adjust.y = 0.5f / phase->output_height;
+	} else if (!destinations.empty()) {
+		assert(destinations.size() == 1);
+		fbo = resource_pool->create_fbo(destinations[0].texnum);
+		glBindFramebuffer(GL_FRAMEBUFFER, fbo);
+		glViewport(0, 0, phase->output_width, phase->output_height);
+	}
+
 	// Give the required parameters to all the effects.
 	unsigned sampler_num = phase->inputs.size();
 	for (unsigned i = 0; i < phase->effects.size(); ++i) {
 		Node *node = phase->effects[i];
 		unsigned old_sampler_num = sampler_num;
-		node->effect->set_gl_state(instance_program_num, phase->effect_ids[node], &sampler_num);
+		node->effect->set_gl_state(instance_program_num, phase->effect_ids[make_pair(node, IN_SAME_PHASE)], &sampler_num);
 		check_error();
 
 		if (node->effect->is_single_texture()) {
@@ -1975,16 +2257,31 @@ void EffectChain::execute_phase(Phase *phase, bool last_phase,
 		}
 	}
 
-	// Uniforms need to come after set_gl_state(), since they can be updated
-	// from there.
-	setup_uniforms(phase);
+	if (phase->is_compute_shader) {
+		unsigned x, y, z;
+		phase->compute_shader_node->effect->get_compute_dimensions(phase->output_width, phase->output_height, &x, &y, &z);
 
-	// Bind the vertex data.
-	GLuint vao = resource_pool->create_vec2_vao(phase->attribute_indexes, vbo);
-	glBindVertexArray(vao);
+		// Uniforms need to come after set_gl_state() _and_ get_compute_dimensions(),
+		// since they can be updated from there.
+		setup_uniforms(phase);
+		glDispatchCompute(x, y, z);
+		check_error();
+		glMemoryBarrier(GL_TEXTURE_FETCH_BARRIER_BIT | GL_TEXTURE_UPDATE_BARRIER_BIT);
+		check_error();
+	} else {
+		// Uniforms need to come after set_gl_state(), since they can be updated
+		// from there.
+		setup_uniforms(phase);
 
-	glDrawArrays(GL_TRIANGLES, 0, 3);
-	check_error();
+		// Bind the vertex data.
+		GLuint vao = resource_pool->create_vec2_vao(phase->attribute_indexes, vbo);
+		glBindVertexArray(vao);
+
+		glDrawArrays(GL_TRIANGLES, 0, 3);
+		check_error();
+
+		resource_pool->release_vec2_vao(vao);
+	}
 	
 	for (unsigned i = 0; i < phase->effects.size(); ++i) {
 		Node *node = phase->effects[i];
@@ -1992,9 +2289,8 @@ void EffectChain::execute_phase(Phase *phase, bool last_phase,
 	}
 
 	resource_pool->unuse_glsl_program(instance_program_num);
-	resource_pool->release_vec2_vao(vao);
 
-	if (!last_phase) {
+	if (fbo != 0) {
 		resource_pool->release_fbo(fbo);
 	}
 }
@@ -2002,6 +2298,12 @@ void EffectChain::execute_phase(Phase *phase, bool last_phase,
 void EffectChain::setup_uniforms(Phase *phase)
 {
 	// TODO: Use UBO blocks.
+	for (size_t i = 0; i < phase->uniforms_image2d.size(); ++i) {
+		const Uniform<int> &uniform = phase->uniforms_image2d[i];
+		if (uniform.location != -1) {
+			glUniform1iv(uniform.location, uniform.num_values, uniform.value);
+		}
+	}
 	for (size_t i = 0; i < phase->uniforms_sampler2d.size(); ++i) {
 		const Uniform<int> &uniform = phase->uniforms_sampler2d[i];
 		if (uniform.location != -1) {
@@ -2021,6 +2323,12 @@ void EffectChain::setup_uniforms(Phase *phase)
 			glUniform1iv(uniform.location, uniform.num_values, uniform.value);
 		}
 	}
+	for (size_t i = 0; i < phase->uniforms_ivec2.size(); ++i) {
+		const Uniform<int> &uniform = phase->uniforms_ivec2[i];
+		if (uniform.location != -1) {
+			glUniform2iv(uniform.location, uniform.num_values, uniform.value);
+		}
+	}
 	for (size_t i = 0; i < phase->uniforms_float.size(); ++i) {
 		const Uniform<float> &uniform = phase->uniforms_float[i];
 		if (uniform.location != -1) {
