@@ -5,6 +5,7 @@
 
 #include <epoxy/gl.h>
 #include <string>
+#include <Eigen/Core>
 
 #include "effect.h"
 
@@ -13,11 +14,12 @@ namespace movit {
 class WhiteBalanceEffect : public Effect {
 public:
 	WhiteBalanceEffect();
-	virtual std::string effect_type_id() const { return "WhiteBalanceEffect"; }
-	virtual AlphaHandling alpha_handling() const { return DONT_CARE_ALPHA_TYPE; }
-	std::string output_fragment_shader();
+	std::string effect_type_id() const override { return "WhiteBalanceEffect"; }
+	AlphaHandling alpha_handling() const override { return DONT_CARE_ALPHA_TYPE; }
+	bool strong_one_to_one_sampling() const override { return true; }
+	std::string output_fragment_shader() override;
 
-	void set_gl_state(GLuint glsl_program_num, const std::string &prefix, unsigned *sampler_num);
+	void set_gl_state(GLuint glsl_program_num, const std::string &prefix, unsigned *sampler_num) override;
 
 private:
 	// The neutral color, in linear sRGB.
@@ -26,6 +28,8 @@ private:
 	// Output color temperature (in Kelvins).
 	// Choosing 6500 will lead to no color cast (ie., the neutral color becomes perfectly gray).
 	float output_color_temperature;
+
+	Eigen::Matrix3d uniform_correction_matrix;
 };
 
 }  // namespace movit
